# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright 2016 Continuum Analytics, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------
from click.testing import CliRunner
import pytest

from .base import VCR
from anaconda_enterprise.cli.main import group


@pytest.mark.integration
def test_create():
    runner = CliRunner()
    with VCR.use_cassette('login.yml'):
        result = runner.invoke(group, ['login'], 'derek\nderek')
        assert result.exit_code == 0


#     with VCR.use_cassette('channels/create.yml'):
#         result = runner.invoke(
#             group,
#             ['channels', 'create', 'test1'], )
#         assert result.output == '''\
# test1
#     '''
#         assert result.exit_code == 0
