

set -ex



pip check
anaconda-enterprise-cli --help
anaconda-enterprise-cli --version
aecli --help
aecli --version
curl -OL https://curl.se/ca/cacert.pem
export CA_CHAIN=$PWD/cacert.pem
py.test tests
exit 0
